#include <Arduino.h>
#include <Wire.h>
#include "Poziomica.h"
#include "vl53l1_api.h"

#define MEASUREMENT_BUDGET_MS 150 
#define INTER_MEASUREMENT_PERIOD_MS 160

VL53L1_Dev_t                   dev;
VL53L1_DEV                     Dev = &dev;

static bool lidarOK;
static bool lidarStopped;

#ifdef USE_GLASS
VL53L1_CalibrationData_t CalibrationData, initCalibrationData;
#endif

int stopLidar(bool forcal, bool forinit)
{
    uint8_t byteData;
    uint16_t wordData;
    VL53L1_software_reset(Dev);
    VL53L1_RdByte(Dev, 0x010F, &byteData);
    printf("VL53L1X Model_ID: %02X\n",byteData);
    VL53L1_RdByte(Dev, 0x0110, &byteData);
    printf("VL53L1X Module_Type: %02X\n",byteData);
    VL53L1_RdWord(Dev, 0x010F, &wordData);
    printf("VL53L1X: %04X\n",wordData);
    int status = VL53L1_WaitDeviceBooted(Dev);
    if (!status) status = VL53L1_DataInit(Dev);
    if (!status) status = VL53L1_StaticInit(Dev);
    if (forcal) {
        lidarStopped=true;
        status = VL53L1_SetPresetMode(Dev, VL53L1_PRESETMODE_LITE_RANGING);
        if (!status) status = VL53L1_SetDistanceMode(Dev, VL53L1_DISTANCEMODE_SHORT);
        if (!status) status = VL53L1_SetMeasurementTimingBudgetMicroSeconds(Dev, 47000UL);
        if (!status) status = VL53L1_SetInterMeasurementPeriodMilliSeconds(Dev,100);
        return 0;
    }
    if (!status) status = VL53L1_SetDistanceMode(Dev, VL53L1_DISTANCEMODE_LONG);
    if (!status) status =
            VL53L1_SetMeasurementTimingBudgetMicroSeconds(Dev,
                                                          (uint32_t)
                                                          MEASUREMENT_BUDGET_MS
                                                          * 1000);
    if (!status) status =
            VL53L1_SetInterMeasurementPeriodMilliSeconds(Dev,
                                                         INTER_MEASUREMENT_PERIOD_MS);

#ifdef USE_GLASS
    if (!status) {
        if (forinit) {
            status=VL53L1_GetCalibrationData(Dev,&initCalibrationData);
            if (!status) {
                if (!preferences.getBytes("calidar",(void *)&CalibrationData,sizeof(CalibrationData))) {
                    CalibrationData = initCalibrationData;
                }
                else {
                    status=VL53L1_SetCalibrationData(Dev,&CalibrationData);
                }
            }
        }
        else {
            status=VL53L1_SetCalibrationData(Dev,&CalibrationData);
        }
    }
#endif

#ifdef HAVE_NARROW        
    if (!status) status=setNarrowAngle(true);
#endif
    if (!status) status = VL53L1_StartMeasurement(Dev);

    if (status) {
        printf("LIDAR Failed\n");
        lidarOK=false;
    }
    else {
        printf("LIDAR OK\n");
        lidarOK=true;
        lidarStopped=false;
    }
    return 0;
}


void initLidar(void)
{
    Dev->I2cDevAddr = 0x52;
    stopLidar(false, true);
}

int calibrateOffset(void)
{
    int status = VL53L1_PerformOffsetSimpleCalibration(Dev, 140);
    return status;
}

int calibrateCross(void)
{
    int status = VL53L1_PerformSingleTargetXTalkCalibration(Dev, crosstalk);
    return status;
}

int calibrateSPAD(void)
{
    int status = VL53L1_PerformRefSpadManagement(Dev);
    return status;
}

int storeCalibration(void)
{
#ifdef USE_GLASS
    int status=VL53L1_GetCalibrationData(Dev,&CalibrationData);
    preferences.putBytes("calidar",(void *)&CalibrationData,sizeof(CalibrationData));
    return status;
#else
    return 0;
#endif
}

void removeLidarCalib(void)
{
#ifdef USE_GLASS
    CalibrationData = initCalibrationData;
    VL53L1_SetCalibrationData(Dev,&CalibrationData);
    preferences.putBytes("calidar",(void *)&CalibrationData,sizeof(CalibrationData));
#endif
}


void printRangingData()
{
  static VL53L1_RangingMeasurementData_t RangingData;

  int status = VL53L1_GetRangingMeasurementData(Dev, &RangingData);
  if(!status)
  {
    Serial.print(RangingData.RangeStatus);
    Serial.print(F(","));
    Serial.print(RangingData.RangeMilliMeter);
    Serial.print(F(","));
    Serial.print(RangingData.SignalRateRtnMegaCps/65536.0);
    Serial.print(F(","));
    Serial.println(RangingData.AmbientRateRtnMegaCps/65336.0);
  }
}

int setLidarDistan(uint8_t distan)
{
    static int dm[]={
        VL53L1_DISTANCEMODE_LONG,
        VL53L1_DISTANCEMODE_MEDIUM,
        VL53L1_DISTANCEMODE_SHORT};
    return VL53L1_SetDistanceMode(Dev, dm[distan]);
}

int setNarrowAngle(bool narrow)
{
#ifdef HAVE_NARROW
    VL53L1_UserRoi_t roiConfig;
    FixPoint1616_t xp, yp;
    int status;
    static int8_t x0=6, y0=6;
    static bool haveOC = false;
    if (!haveOC) {
        status = VL53L1_GetOpticalCenter(Dev, &xp, &yp);
        if (status) status = VL53L1_GetOpticalCenter(Dev, &xp, &yp);
        if (status) return status;
        x0 = ((xp >> 16) & 15)-1;
        y0 = ((yp >> 16) & 15)-1;
        x0 = constrain(x0, 0,12);
        y0 = constrain(y0, 0,12);
        haveOC=true;
    }
    roiConfig.TopLeftX  = narrow?x0:0;
    roiConfig.TopLeftY  = narrow?y0+3:15;
    roiConfig.BotRightX = narrow?x0+3:15;
    roiConfig.BotRightY = narrow?y0:0;
    
    return VL53L1_SetUserROI(Dev, &roiConfig);
#else
    return 0;
#endif
}


#define DRB_COUNT 9
static int16_t distanrb[DRB_COUNT];
static int drbLen=0, drbPos=0;

static void setCurrentDistance(int distance)
{
    if (distance >= 0) {
        distanrb[drbPos] = distance;
        drbPos = (drbPos + 1) % DRB_COUNT;
        if (drbLen < DRB_COUNT) drbLen++;
    }
    else {
        drbLen=0;
    }
}

static int distance;
static uint32_t distance_ms;
int getDistance(void)
{
    int16_t distances[DRB_COUNT];
    int i,j,k,k2,d;

    if (!drbLen || millis() - distance_ms >1500UL) {
        return -2;
    }
    for (i=0, j=drbPos; i< drbLen; i++) {
        j--;
        if (j < 0) j=DRB_COUNT -1;
        d=distanrb[j];
        if (!i) distances[0] = d;
        else {
            for (k=0; k<i;k++) if (d < distances[k]) break;
            for (k2=i;k2>k; k2--) distances[k2] = distances[k2-1];
            distances[k] = d;
        }
    }
    if (i == 1) return distances[0];
    if (i<4) {
        k=0; k2=i;
    }
    else {
        k=1; k2=i-1;
    }
    for (d=0,j=k; j<k2; j++) {
        d += distances[j];
    }
    return d/(k2-k);
}

void lidarLoop(void)
{
    uint8_t isReady;
    int status;
    static uint32_t startMs = millis();
    static VL53L1_RangingMeasurementData_t RangingData;
    if (!lidarOK || lidarStopped) return;
    status = VL53L1_GetMeasurementDataReady(Dev, &isReady);
    if (!status) {
        if (isReady) {
            status = VL53L1_GetRangingMeasurementData(Dev, &RangingData);
            if (!status) {
                status = RangingData.RangeStatus;
                distance = RangingData.RangeMilliMeter;
                if (distance > 4000 || status > 7) distance = -2;
                distance_ms = millis();
                setCurrentDistance(distance);
            }
            VL53L1_ClearInterruptAndStartMeasurement(Dev);
            startMs = millis();
        } else if (millis() - startMs >
                   VL53L1_RANGE_COMPLETION_POLLING_TIMEOUT_MS) {
            VL53L1_ClearInterruptAndStartMeasurement(Dev);
            startMs = millis();
        }
    } else {
        printf("Error getting data ready: %d\n",status);
    }
}

